#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>
#include "StdFps.h"

//Global variables
int MaxIterations=200;
double XMin=-2, XMax=1, YMin=-1, YMax=1; //The X/Y axis min/max (boundary) points
DWORD Width=600, Height=400;

//Vertex structure
struct Vertex
{
    FLOAT x, y, z;
    FLOAT u, v;
};

//The D3D Device
IDirect3DDevice9* g_pd3dDevice;

//Function list
char* Main(HINSTANCE hInst);
char* CreateReleaseD3D(HWND hWnd, bool Creating); //Create and Release D3D Interface
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam); //Process windows events
void DrawToWindow(); //Runs a render pass

//Program entry point
int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR pCmdLine, int nCmdShow)
{
	char* MainReturn=Main(hInst);
	if(MainReturn)
	{
		MessageBox(NULL, MainReturn, "Error", MB_OK);
		return S_FALSE;
	}

	return S_OK;
}

//Main function
char* Main(HINSTANCE hInst)
{
	//Create the window class
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), NULL, WndProc, 0L, 0L, hInst, LoadIcon(hInst, "MAIN_ICON"), LoadCursor(NULL,IDC_ARROW), NULL, NULL, "JuliaFractal", NULL };
	if(RegisterClassEx(&wc)==0)
		return "Cannot register windows class";

	//Determine the size of the actual window that we want based upon client area dimensions
	RECT WindowRect={0, 0, Width, Height};
	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, false);

	//Create the window and display it
	HWND hWnd=CreateWindowEx(NULL, "JuliaFractal", "Julia Fractal Pixel Shader Test", WS_OVERLAPPEDWINDOW, 0, 0, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInst, NULL);
	
	//Create D3D Interface
	char* D3DResult=CreateReleaseD3D(hWnd, true);
	if(D3DResult)
		return D3DResult;

	//Show the window
	ShowWindow(hWnd, SW_SHOW);
	
	//Main loop listening for events
	FPS MyFPS;
	MSG msg;
	ZeroMemory(&msg, sizeof(MSG));
	while(true)
	{
		//Check for a windows message
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if(GetMessage(&msg, NULL, 0, 0)==0)
				break;
			TranslateMessage(&msg); 
			DispatchMessage(&msg);
		}

		//Check if a mouse button is being pressed...
		int MouseStates=(GetAsyncKeyState(VK_LBUTTON) ? 1 : 0)|(GetAsyncKeyState(VK_RBUTTON) ? 2 : 0);
		if(!MouseStates)
			continue;

		//Make sure we are in the window boundaries
		POINT CursorPos;
		GetCursorPos(&CursorPos);
		ScreenToClient(hWnd, &CursorPos);
		if(CursorPos.x<0 || CursorPos.y<0 || CursorPos.x>=(int)Width || CursorPos.y>=(int)Height)
			continue;

		//Zoom in/out around point
		double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window dimensions
		double PointPercentX=(double)CursorPos.x/Width, PointPercentY=(double)CursorPos.y/Height;
		double PointX=XMin+PointPercentX*ScaleX, PointY=YMin+PointPercentY*ScaleY;
		double ScaleNum=(MouseStates&1 ? 1.0f/1.03 : 1.03);
		
		XMin=PointX-(ScaleX*ScaleNum)*(PointPercentX);
		XMax=PointX+(ScaleX*ScaleNum)*(1-PointPercentX);
		YMin=PointY-(ScaleY*ScaleNum)*(PointPercentY);
		YMax=PointY+(ScaleY*ScaleNum)*(1-PointPercentY);

		float PSFConstants[]={(float)(XMax-XMin), (float)(YMax-YMin), (float)YMin, (float)XMin}; //XWidth=XMax-XMin, YHeight=YMax-YMin, YMin, XMin
		g_pd3dDevice->SetPixelShaderConstantF(0, PSFConstants, 1);
		DrawToWindow();
		
		//Update the window title with the current FPS
		MyFPS.AddFrame();
		char WindowTitle[100]="Julia Fractal @ ";
		strcat(WindowTitle, MyFPS.GetFpsString());
		SetWindowText(hWnd, WindowTitle);

	}

	//Release D3D Devices
	CreateReleaseD3D(hWnd, false);

    return NULL;
}

char* CreateReleaseD3D(HWND hWnd, bool Creating) //Create and Release D3D Interface
{
	static IDirect3D9* D3DInterface;
	static IDirect3DVertexBuffer9* VertextBuffer;

	//Release the device if requested
	if(!Creating)
	{
		if(VertextBuffer)
			VertextBuffer->Release();
		if(g_pd3dDevice)
			g_pd3dDevice->Release();
		if(D3DInterface)
			D3DInterface->Release();
		return NULL;
	}

	//Create the DX interface
    D3DInterface=Direct3DCreate9(D3D_SDK_VERSION);
    if(!D3DInterface)
		return "Cannot create DirectX Interface"; 

	//Determine if we have the needed capabilities, and if not, set to render via software
    D3DDEVTYPE dev_type=D3DDEVTYPE_HAL;
    D3DCAPS9 d3dcaps9;
    ZeroMemory(&d3dcaps9, sizeof(d3dcaps9));
    if(FAILED(D3DInterface->GetDeviceCaps(D3DADAPTER_DEFAULT, dev_type, &d3dcaps9)) || d3dcaps9.PixelShaderVersion<D3DPS_VERSION(2,1))
		dev_type=D3DDEVTYPE_REF;

    //Get the current display mode
    D3DDISPLAYMODE displayMode;
    ZeroMemory(&displayMode, sizeof(displayMode));
    if(FAILED(D3DInterface->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &displayMode)))
		return "Cannot query display mode";

	//Create the DX Device
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory(&d3dpp, sizeof(d3dpp));
    d3dpp.BackBufferFormat=displayMode.Format;
    d3dpp.Windowed=true;
    d3dpp.SwapEffect=D3DSWAPEFFECT_DISCARD;
    d3dpp.EnableAutoDepthStencil=FALSE;
    if(FAILED(D3DInterface->CreateDevice(D3DADAPTER_DEFAULT, dev_type, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &g_pd3dDevice)))
		return "Cannot create DirectX Device";

	//Define a full screen quad
	const int VERTEX_FORMAT=(D3DFVF_XYZ|D3DFVF_TEX1);
	float HWidth=(float)Width/2, HHeight=(float)Height/2;
	Vertex Vertices[4] = 
	{
		-HWidth, HHeight,1,0,0,
		 HWidth, HHeight,1,1,0,
		 HWidth,-HHeight,1,1,1,
		-HWidth,-HHeight,1,0,1
	};

	//Create the vertex buffer
    void* ptr;
    if(
		FAILED(g_pd3dDevice->CreateVertexBuffer(4*sizeof(Vertex), D3DUSAGE_WRITEONLY, VERTEX_FORMAT, D3DPOOL_DEFAULT, &VertextBuffer, NULL)) ||
		FAILED(VertextBuffer->Lock(0, 4*sizeof(Vertex), &ptr, 0))
	)
		return "Cannot create vertex buffer";
    memcpy(ptr, Vertices, 4*sizeof(Vertex));
    if(
		FAILED(VertextBuffer->Unlock()) ||
		FAILED(g_pd3dDevice->SetFVF(VERTEX_FORMAT)) ||
		FAILED(g_pd3dDevice->SetStreamSource(0, VertextBuffer, 0, sizeof(Vertex)))
	)
		return "Cannot populate vertex buffer";

    //Create and set the pixel shader
    LPD3DXBUFFER pCode, pErrorMsgs=NULL;
	IDirect3DPixelShader9* PixelShader;
    if(FAILED(D3DXAssembleShaderFromFile("JuliaFractalPixelShader.ps", NULL, NULL, 0, &pCode, &pErrorMsgs)))
	{
		if(!pErrorMsgs)
			return "Cannot find pixel shader file";
		MessageBox(NULL, (char*)pErrorMsgs->GetBufferPointer(), "Pixel Shader Error", MB_OK);
        return "Cannot create pixel shader";
	}
    if(
		FAILED(g_pd3dDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), &PixelShader)) ||
		FAILED(g_pd3dDevice->SetPixelShader(PixelShader))
	)
		return "Cannot initate pixel shader";
    PixelShader->Release();

    //Render states
    g_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE); //No lighting

	//Prepare a 2D screen
	D3DXMATRIX Ortho2D;	
	D3DXMATRIX Identity;
	D3DXMatrixOrthoLH(&Ortho2D, (float)Width, (float)Height, 0.0f, 1.0f);
	D3DXMatrixIdentity(&Identity);
	g_pd3dDevice->SetTransform(D3DTS_PROJECTION, &Ortho2D);
	g_pd3dDevice->SetTransform(D3DTS_WORLD, &Identity);
	g_pd3dDevice->SetTransform(D3DTS_VIEW, &Identity);

    //Create the color range palette as a bitmap
	const DWORD ColorRanges[]={
		0x000080, //Dark Blue
		0x0000FF, //Blue
		0xA0A0FF, //Light Blue
		0xFFFF80, //Light yellow
		0x80FF80, //Light green
		0x008000, //Dark green
		0x008080, //Forest green
		0x00FFFF, //Cyan
		0xFF00FF, //Purple
		0xFF0000, //Red
		0x800000, //Dark red
		0x000080 //Dark blue - Start back at the beginning
	}, NumColorRanges=sizeof(ColorRanges)/sizeof(DWORD);
	const int FHS=sizeof(BITMAPFILEHEADER),IHS=sizeof(BITMAPINFOHEADER); //File header and info header sizes
	BITMAPFILEHEADER BFH={0x4D42, FHS+IHS+sizeof(ColorRanges), 0, 0, FHS+IHS};
	BITMAPINFOHEADER BIH={IHS, NumColorRanges, 1, 1, 32, BI_RGB,	sizeof(ColorRanges), 0, 0, 0, 0};
	BYTE *MyPixelBitmap=new BYTE[FHS+IHS+sizeof(ColorRanges)];
	memcpy(MyPixelBitmap, &BFH, FHS);
	memcpy(MyPixelBitmap+FHS, &BIH, IHS);
	memcpy(MyPixelBitmap+FHS+IHS, ColorRanges, sizeof(ColorRanges));

	//Copy the bitmap to a texture
	IDirect3DTexture9* PaletteTexture;
	if(
		FAILED(D3DXCreateTextureFromFileInMemoryEx(g_pd3dDevice, MyPixelBitmap, FHS+IHS+sizeof(ColorRanges), NumColorRanges, 1, 1, 0, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, D3DX_FILTER_LINEAR, D3DX_FILTER_LINEAR, 0, NULL, NULL, &PaletteTexture)) ||
		FAILED(g_pd3dDevice->SetTexture(0, PaletteTexture)) ||
		FAILED(g_pd3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR)) ||
		FAILED(g_pd3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR))
	)
		return "Cannot create palette";
    PaletteTexture->Release();
	delete[] MyPixelBitmap;

	//Set the pixel shader constants
	const int RangeSize=20;
	const float PSFConstants[]={
		(float)(XMax-XMin), (float)(YMax-YMin), (float)YMin, (float)XMin, //XWidth=XMax-XMin, YHeight=YMax-YMin, YMin, XMin
		1,0,1,-1, //Dot product multiply for xtemp
		0,1,0,2, //Dot product multiply for y
		1,1,0,(float)MaxIterations, //xyz=Dot product for Complex#^2 breakout test, w=MaxIterations (EscapeLevel)
		4,1.0f/(RangeSize*NumColorRanges),0.32663425997828f,-1.4426950408889f, //BailoutRadius^2, ClampToTexture=1/(RangeSize*NumColorRanges), log(2*log(BailoutRadius)) [used for escape lerp], -1/log(2) [used for escape lerp]
		0.6931471806f,0,0,0 //ReverseLogBase2=ln(2)
	};
	const int PSIConstants[]={MaxIterations,0,0,0}; //MaxIterations,0,0,0
	g_pd3dDevice->SetPixelShaderConstantF(0, PSFConstants, 6);
	g_pd3dDevice->SetPixelShaderConstantI(0, PSIConstants, 1);

	return NULL;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) //Process windows events
{
    switch(msg)
    {
		case WM_SIZE:
		{
			//Store the new client dimensions
			Width=LOWORD(lParam);
			Height=HIWORD(lParam);

			//Release and Create the D3D Interface
			CreateReleaseD3D(hWnd, false);
			CreateReleaseD3D(hWnd, true);
			DrawToWindow();
			break;
		}
		case WM_CHAR:
			if(wParam=='=' || wParam=='-') //Detail level up
			{
				if(wParam=='=')
					MaxIterations*=2;
				else
					MaxIterations/=2;

				const float PSFConstants[]={1,1,0,(float)MaxIterations};
				const int PSIConstants[]={MaxIterations,0,0,0}; //MaxIterations,0,0,0
				g_pd3dDevice->SetPixelShaderConstantF(3, PSFConstants, 1);
				g_pd3dDevice->SetPixelShaderConstantI(0, PSIConstants, 1);

				DrawToWindow();
			}
			else //Assumes a direction key was pressed
			{
				//Move in given direction
				double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window dimensions

				//Determine the center of the window
				double PointX=.5*ScaleX+XMin;
				double PointY=YMax-.5*ScaleY;

				if(wParam=='w')
					PointY+=ScaleY/-4;
				else if(wParam=='a')
					PointX+=ScaleX/-4;
				else if(wParam=='s')
					PointY+=ScaleY/4;
				else if(wParam=='d')
					PointX+=ScaleX/4;
				else //If not w/a/s/d, do nothing
					break;

				XMin=PointX-ScaleX/2;
				XMax=PointX+ScaleX/2;
				YMin=PointY-ScaleY/2;
				YMax=PointY+ScaleY/2;

				float PSFConstants[]={(float)(XMax-XMin), (float)(YMax-YMin), (float)YMin, (float)XMin}; //XWidth=XMax-XMin, YHeight=YMax-YMin, YMin, XMin
				g_pd3dDevice->SetPixelShaderConstantF(0, PSFConstants, 1);

				DrawToWindow();
			}
			break;
		case WM_PAINT:
			//Refresh window
			DrawToWindow();
			break;
		case WM_DESTROY:
			//End the program
			PostQuitMessage(0);
			return 0;
    }
    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void DrawToWindow() //Runs a render pass
{
	g_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_COLORVALUE(1.0, 1.0, 1.0, 1.0), 1.0, 0);
	g_pd3dDevice->BeginScene();
	g_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
	g_pd3dDevice->EndScene();
	g_pd3dDevice->Present(NULL, NULL, NULL, NULL);
}